// Nathan E. Moore
// March 2, 2005

#include "common.h"
#include "idlist.h"
#include "taskalloc.h"
#include <cstddef>
#include <memory.h>
#include <shlobj.h>

namespace com
{
	namespace idlist
	{
		namespace
		{
			const ITEMIDLIST* Next(const ITEMIDLIST* pidl)
			{
				return reinterpret_cast<const ITEMIDLIST*>(reinterpret_cast<const char*>(pidl) + pidl->mkid.cb);
			}

			std::size_t SizeOf(const ITEMIDLIST* pidl)
			{
				// a null pointer has no size
				ASSERTION(pidl);

				std::size_t size = 2;	// start at 2 for the null
				const ITEMIDLIST* current = pidl;
				while (current->mkid.cb)
				{
					size += current->mkid.cb;
					current = Next(current);
				}
				return size;
			}

			const ITEMIDLIST* LastElementWorker(const ITEMIDLIST* pidl)
			{
				ASSERTION(pidl);
				const ITEMIDLIST* file = pidl;
				const ITEMIDLIST* end = file;
				while (end->mkid.cb)
				{
					file = end;
					end = Next(end);
				}
				return file;
			}
		}

		// For all of these function you must free the return ITEMIDLIST with the COM allocator
		ITEMIDLIST* ContainingElement(const ITEMIDLIST* pidl)
		{		
			size_t size = SizeOf(pidl) - SizeOf(LastElementWorker(pidl)) + 2;
			ITEMIDLIST* newidlist(static_cast<ITEMIDLIST*>(TaskAlloc::Alloc(size)));
			memcpy(newidlist, pidl, size - 2);
			*(reinterpret_cast<USHORT*>((reinterpret_cast<char*>(newidlist) + size - 2))) = 0;
			return newidlist;
		}

		ITEMIDLIST* LastElement(const ITEMIDLIST* pidl)
		{
			return Clone(LastElementWorker(pidl));
		}

		ITEMIDLIST* Clone(const ITEMIDLIST* pidl)
		{
			ASSERTION(pidl);

			std::size_t length(SizeOf(pidl));
			ITEMIDLIST* idlist(static_cast<ITEMIDLIST*>(TaskAlloc::Alloc(length)));
			memcpy(idlist, pidl, length);
			return idlist;
		}

		bool Equal(const ITEMIDLIST* lhs, const ITEMIDLIST* rhs)
		{
			std::size_t lsize(SizeOf(lhs));
			std::size_t rsize(SizeOf(rhs));

			if (lsize != rsize)
				return false;
			else
				return 0 == memcmp(lhs, rhs, lsize);
		}

		bool IsDesktop(const ITEMIDLIST* lhs)
		{
			return 0 == lhs->mkid.cb;
		}
	}	//end namespace idlist
}	// end namespace com
